prompt_list = ['deep ocean', 'deep ocean', 'a cute cat', 'a cute cat', 'a naughty dog', 'a cute cat', 'deep ocean', 'dive into pool', 'blue sky', 'a naughty dog', 'dive into pool', 'blue sky', 'a naughty dog', 'deep ocean', 'deep ocean', 'a naughty dog', 'dive into pool', 'deep ocean', 'a cute cat', 'blue sky', 'blue sky', 'dive into pool', 'a cute cat', 'blue sky', 'deep ocean', 'a naughty dog', 'deep ocean', 'dive into pool', 'blue sky', 'deep ocean', 'dive into pool', 'deep ocean', 'blue sky', 'dive into pool', 'a cute cat', 'a cute cat', 'blue sky', 'a cute cat', 'blue sky', 'a cute cat', 'a naughty dog', 'dive into pool', 'blue sky', 'a naughty dog', 'blue sky', 'dive into pool', 'deep ocean', 'deep ocean', 'dive into pool', 'a cute cat', 'deep ocean',
               'a naughty dog', 'deep ocean', 'a cute cat', 'deep ocean', 'a cute cat', 'a naughty dog', 'a cute cat', 'a cute cat', 'dive into pool', 'deep ocean', 'a cute cat', 'dive into pool', 'a naughty dog', 'dive into pool', 'a cute cat', 'blue sky', 'deep ocean', 'blue sky', 'a naughty dog', 'deep ocean', 'a naughty dog', 'dive into pool', 'dive into pool', 'a naughty dog', 'a naughty dog', 'a naughty dog', 'a naughty dog', 'blue sky', 'a cute cat', 'a cute cat', 'a cute cat', 'dive into pool', 'dive into pool', 'dive into pool', 'a cute cat', 'a cute cat', 'a naughty dog', 'dive into pool', 'a cute cat', 'blue sky', 'a cute cat', 'dive into pool', 'dive into pool', 'dive into pool', 'blue sky', 'a cute cat', 'a cute cat', 'a naughty dog', 'blue sky']
