#include "omniback/helper/toml_parser.hpp"
#include <system_error>

namespace om::toml {
str::mapmap parse(const std::string& toml_str) {
  throw std::runtime_error("Not implemented");
  return str::mapmap();
}
} // namespace om::toml
